/*
* Simple Linear Regression - Tristan Aubrey-Jones 2014
* Performs simple linear regressionin parallel on 
* 2D coordinates stored as doubles in a binary input file.
*/
//#include <vector>
//#include <cstdlib>
//#include <memory>
//#include <cstring> // for strcmp
//#include <cstdio> // sprintf

#include <iostream>
#include <fstream>
#include <mpi.h>

using namespace std;

typedef struct InStream {
  ifstream *fstream; // file input stream
  char *readBuf, *bufPos; // buffer, and current buffer position
  int blockSize; // num of chars in a block
  int readBufSize; // number of chars in the buffer
  int blockCount; // current num of blocks in buffer
  int blockNum; // the num of blocks from the buffer that have been read so far

  InStream(int blockSz, ifstream *s): fstream(s), blockSize(blockSz) {
    readBufSize = 1024*1024*blockSz; 
    readBuf = new char[readBufSize];
    bufPos = readBuf;
    blockNum = 0; // no blocks in buffer to start
    blockCount = 0; // none in buffer, forces a read when first calls readBlock
  }

  ~InStream() {
     delete [] readBuf;
   }

  char* readBlock() {
    // more in buffer
    if (blockNum < blockCount) {
      // return block in buffer
      char* result = bufPos;
      bufPos += blockSize;
      blockNum++;
      return result;
    }
    // need to read in new block from file 
    else {
      // if not eof
      if (fstream->eof()) {
        return NULL;
      } else {
        // read in block from file
        fstream->read(readBuf, readBufSize);
        blockCount = fstream->gcount() / blockSize;
        // point to second element in buffer:
        blockNum = 1;
        bufPos = readBuf + blockSize;
        // return pointer to first element in buffer
        return readBuf;
      }
    }
  };

} InStream;

typedef unsigned long ulong;

// simple linear regression 
// for binary file of 2D double points
int main(int argc, char** argv) {
  // init
  MPI::Init(argc, argv);
  int rank = MPI::COMM_WORLD.Get_rank();
  int N = MPI::COMM_WORLD.Get_size();

  // input file
  const char *path = "file1.bin";

  // open file
  streampos begin,end;
  ifstream myfile (path, ios::binary);
  if (!myfile.is_open()) {
    cout << "Could not open file " << path << "!\n";
    return -1;
  }

  // get file size
  begin = myfile.tellg();
  myfile.seekg (0, ios::end);
  end = myfile.tellg();
  ulong fileSize = end-begin;
  int blockSize = sizeof(double)*2;
  ulong totalBlocks = fileSize / blockSize;
  if (fileSize % blockSize != 0) {
    cerr << "Filesize " << fileSize << " is not a multiple of " << blockSize << "!\n";
    throw -1;
  }
  
  // get portion of file to read
  ulong numBlocks = totalBlocks / N;
  ulong startBlock = rank * numBlocks;
  if (rank == N-1) numBlocks += totalBlocks - (N * numBlocks);

  // seek to start
  myfile.seekg(startBlock*blockSize, ios::beg);
  
  // init input stream
  InStream stream1(blockSize/2, &myfile);
  
  // read once, summing data
  double totalX = 0.0, totalY = 0.0;
  for (ulong i = 0; i < numBlocks; i++) {
    double x = *stream1.readBlock();
    double y = *stream1.readBlock();
    totalX += x; totalY += y;
  }

  // all reduce totals
  double locTotals[2], totals[2];
  locTotals[0] = totalX;
  locTotals[1] = totalY;
  MPI::COMM_WORLD.Allreduce(&locTotals[0], &totals[0], 2, MPI_DOUBLE, MPI_SUM);
  totalX = totals[0];
  totalY = totals[1];

  // calc means
  double avgX = totalX / numBlocks;
  double avgY = totalY / numBlocks;

  // seek to start
  myfile.seekg(startBlock*blockSize, ios::beg);
  
  // init input stream
  InStream stream2(blockSize/2, &myfile);

  // read again, calculating top and bottom of frac
  double top, bot;
  for (ulong i = 0; i < numBlocks; i++) {
    double x = *stream1.readBlock();
    double y = *stream1.readBlock();
    top += (x - avgX)*(y - avgY);
    bot += (x - avgX)*(x - avgX);
  }

  // close file
  myfile.close();

  // gather tops and bots to root
  double sendbuf[2], recvbuf[2];
  sendbuf[0] = top;
  sendbuf[1] = bot;
  MPI::COMM_WORLD.Gather(&sendbuf[0], 2, MPI_DOUBLE,
    &recvbuf[0], 2, MPI_DOUBLE, 0);

  if (rank == 0) {
    // unpack result
    top = recvbuf[0];
    bot = recvbuf[1];

    // calc coeffs
    double beta = top / bot;
    double alpha = avgY - (avgX * beta);

    // display
    cout << "y = " << alpha << " + " << beta << "x\n";
  }

  // end
  MPI::Finalize();
  return 0;
}

